import os
import sys
from pathlib import Path

cur_file = os.path.abspath(__file__)
cur_file = Path(cur_file)
sys.path.insert(0, cur_file.parent.as_posix())
sys.path.insert(0, cur_file.parent.parent.as_posix())

import argparse
import subprocess
from fact_table_ana import pred_baseq_and_emp_q, rq_iy_analysis


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-q", type=str, required=True, help="query file")
    parser.add_argument("-t", type=str, required=True, help="target file")
    parser.add_argument("-f", action="store_true", help="force to overwrite the output bam file")
    parser.add_argument("--qname-suffix", type=str, default=None, help="only the qname endswith $qname_suffix will be considered", dest="qname_suffix")

    args = parser.parse_args()

    """
    gsmm2 align -q $query_file -t $ref_file -p outputbam_prefix --noMar
    gsetl --outdir $outdir aligned-bam --bam $aligned_bam --ref-file $ref_file
    """

    file_stem = Path(args.q).stem

    oup_bam_prefix = f"{args.q}.gsmm2-aligned"
    oup_bam = f"{oup_bam_prefix}.bam"
    if args.f and os.path.exists(oup_bam):
        os.remove(oup_bam)
    if not os.path.exists(oup_bam):
        ## 1. do alignment
        cmd_str = f"gsmm2 align -q {args.q} -t {args.t} -p {oup_bam_prefix} --noMar"
        if args.qname_suffix:
            cmd_str += f" --qname-suffix {args.qname_suffix}"
        subprocess.check_call(cmd_str, shell=True)

    ## 2. gsetl extract fact table
    gsetl_outdir = f"{oup_bam}-gsetl"
    cmd_str = f"gsetl -f --outdir {gsetl_outdir} aligned-bam --bam {oup_bam} --ref-file {args.t} --factPolyInfo 0 "
    subprocess.check_call(cmd_str, shell=True)

    ## 3. do analysis

    baseq_ana_args = {
        "fact_table": f"{gsetl_outdir}/fact_baseq_stat.csv",
        "o_prefix": file_stem,
    }
    pred_baseq_and_emp_q.main(argparse.Namespace(**baseq_ana_args))

    readsq_ana_args = {
        "fact_table": f"{gsetl_outdir}/fact_aligned_bam_bam_basic.csv",
        "o_prefix": file_stem,
    }
    rq_iy_analysis.main(argparse.Namespace(**readsq_ana_args))


if __name__ == "__main__":
    main()



import mappy as mp
import pysam
from collections import defaultdict
from tqdm import tqdm
from typing import List, Tuple

def calculate_identity(cigar_tuple):
    """
    Calculate the identity of a read based on CIGAR tuple with 'eqx' operations.

    Args:
        cigar_tuple (list): A list of (operation, length) tuples from a CIGAR string.

    Returns:
        float: The calculated identity (matches / alignment length).
    """
    matches = 0
    alignment_length = 0
    
    for length, op in cigar_tuple:
        if op == 7:  # Exact match (EQ)
            matches += length
            alignment_length += length
        elif op == 8:  # Mismatch (X)
            alignment_length += length
        elif op in {0, 1, 2}:  # Match (M), Insertion (I), Deletion (D)
            alignment_length += length

    if alignment_length == 0:
        return 0.0
    return matches / alignment_length

def pin_start_end(cigar_tuple: List[Tuple[int, int]]):
    if len(cigar_tuple)>0:
        if cigar_tuple[0][1] != 7:
            cigar_tuple.pop(0)
    if len(cigar_tuple) >0:
        if cigar_tuple[-1][1] != 7:
            cigar_tuple.pop(-1)
    return cigar_tuple

def analyze_bam(bam_file: str):
    """
    Analyze a BAM file for strand inconsistency based on 'channel' tag.

    Args:
        bam_file (str): Path to the input BAM file.
    """
    # Open the BAM file
    bam = pysam.AlignmentFile(bam_file, "rb", check_sq=False, threads=40)

    # Group reads by 'channel' tag
    reads_by_channel = defaultdict(lambda: {'fwd': None, 'rev': None})

    channels = set()
    
    for read in tqdm(bam.fetch(until_eof=True), desc=f"reading {bam_file}"):
        channel = read.get_tag('channel')
        name = read.query_name
        channels.add(channel)
        if name.endswith('fwd'):
            reads_by_channel[channel]['fwd'] = read.query_sequence
        elif name.endswith('rev'):
            reads_by_channel[channel]['rev'] = read.query_sequence

    bam.close()

    # Align and check inconsistencies
    # inconsistent_reads = []
    fwd_rev_channels = 0
    inconsistent_channels = 0
    for channel, reads in tqdm(reads_by_channel.items(), desc='Analyzing channels'):
        if reads['fwd'] is not None and reads['rev'] is not None:
            fwd_rev_channels += 1
            fwd_seq = reads['fwd']
            rev_seq = reads['rev']

            # Use mappy to align
            aligner = mp.Aligner(seq=fwd_seq, extra_flags = 67108864)  # Default to nucleotide alignment
            
            for hit in aligner.map(rev_seq):
                # print(hit.cigar)
                # print(hit.NM)
                cigar = pin_start_end(hit.cigar)
                identity = calculate_identity(cigar)
                if 0.99 < identity < 0.999:
                    inconsistent_channels += 1
    print("tot:{}, fwd_rev_channels:{}, inconsistent_channels:{}, inconsistent_ratio:{}".format(len(channels), fwd_rev_channels, inconsistent_channels, inconsistent_channels/fwd_rev_channels))

    # Print inconsistent reads
    # if inconsistent_reads:
    #     print('Inconsistent reads found:')
    #     for channel, fwd_name, rev_name in inconsistent_reads:
    #         print(f'Channel: {channel}, Fwd: {fwd_name}, Rev: {rev_name}')
    # else:
    #     print('No inconsistencies found.')
        
    
def main():
    analyze_bam("/data/ccs_data/case-study/20250310-lowQ30/output-bystrand.smc_all_reads.bam")        

if __name__ == "__main__":
    main()